from sqlalchemy import *

from dataProviderBase import *
from T0Mon.utils import *
from T0Mon.analysis import *
from T0Mon.consts import *
from T0Mon.sqlManager import *

from T0Mon.globals import logger
import T0Mon.config

### Module for classes providing data about T0 runs ###

class RunConfigKeyProvider(TableDataProvider):
    col1 = [Col("Dataset"), Cols([])]
    col2 = [Col("Repack"), Cols([Col("Proc Version")])]
    col3 = [Col("Reco"), Cols([Col("Reco"), Col("CMSsw Version"), Col("Global tag"), Col("Config URL", decorator=buildWebLink), Col("Proc Version")])]
    col4 = [Col("AlcaSkim"), Cols([Col("AlcaSkim"), Col("CMSsw Version"), Col("Config URL", decorator=buildWebLink), Col("Proc Version")])]
    columns = []
    columns.append(col1)
    columns.append(col2)
    columns.append(col3)
    columns.append(col4)

    cols = DoubleCols(columns)

    def __init__(self, runid):
        self.runid = runid
        self.title = "Run " + str(self.runid)
        TableDataProvider.__init__(self)
        TableDataProvider._load(self)

    def _buildQuery(self):
        repackConfig = Tbl(TABLES.REPACK_CONFIG, True)
        recoConfig = Tbl(TABLES.RECO_CONFIG, True)
	alcaConfig = Tbl(TABLES.ALCA_CONFIG, True)
        primaryDataset = Tbl(TABLES.PRIMARY_DATASET, True)
        cmsswVersion = Tbl(TABLES.CMSSW_VERSION, True)
        
        if self.asc:
            order = "ASC"
        else:
            order = "DESC"


        self.query ="""
        SELECT sub.ds_dataset, sub.rp_proc_version, case sub.ro_reco when 0 then 'Disabled' when 1 then 'Enabled' end ro_reco, sub.ro_cmssw_version, sub.ro_global_tag, sub.ro_config_url, sub.ro_proc_version, case sub.al_alcaskim when 0 then 'Disabled' when 1 then 'Enabled' end al_alcaskim, sub.al_cmssw_version, sub.al_config_url, sub.al_proc_version 
        FROM(
               SELECT
               ds.name ds_dataset,
               rp.proc_version rp_proc_version,
               ro.do_reco ro_reco,
               sw1.name ro_cmssw_version,
               ro.global_tag ro_global_tag,
               ro.config_url ro_config_url,
               ro.proc_version ro_proc_version,
	       al.do_alca al_alcaskim,
	       sw2.name al_cmssw_version,
	       al.config_url al_config_url,
	       al.proc_version al_proc_version
               FROM """ + repackConfig + """ rp, """ + recoConfig + """ ro, """ + alcaConfig  + """ al, """  + cmsswVersion  + """ sw1, """ + cmsswVersion  + """ sw2, """ + primaryDataset  + """ ds
               WHERE rp.run_id = """ + str(self.runid) + """
               AND rp.run_id = ro.run_id AND ro.run_id = al.run_id  AND rp.primary_dataset_id = ro.primary_dataset_id AND ro.primary_dataset_id = al.primary_dataset_id AND ro.cmssw_version_id = sw1.id AND al.cmssw_version_id = sw2.id  AND rp.primary_dataset_id = ds.id
               ORDER BY """ + str(int(self.orderby)+1) + " " + order + """
        ) sub
        """
        

class RunsDataProvider(PagedTableDataProvider):
    RUN_ID_TITLE = "Run ID"
    STREAMERS_TITLE = "Streamers"
    REPACKED_TITLE = "Repacked"
    RECO_TITLE = "Reconstructed"
    ALCASKIM_TITLE = "AlcaSkimmed"

    cols = Cols([
        Col(RUN_ID_TITLE, decorator=formRunLink),
        Col("Start", decorator=timeToClock),
        Col("End", decorator=timeToClock),
        Col("Process"),
        Col("Version"),
        Col("Express"),
        Col("Repack"),
        Col("HLTKey", decorator=buildLeftStrip),
        Col("ACQ_era"),
        Col("Status", info=
            """
                <i style="color:blue;font-size:10px">Active</i> - Run is ongoing and streamer data is still arriving for it<br />
                <i style="color:blue;font-size:10px">CloseOutSegmentInjector</i> - The Tier-0 is trying to make consistent luminosity sections to be repacked (currently we need 4*n streamer files per Luminosity Section)<br />
                <i style="color:blue;font-size:10px">CloseOutScheduler</i> - There is streamer data for the run that has not been assigned to a repack job yet<br />
                <i style="color:blue;font-size:10px">CloseOutRepack</i> - There is streamer data in a run that has not been repacked yet (create repack jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutRepackMerge</i> - There is RAW data in the run that needs to be merged (create repack merge jobs and wait from them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutPromptReco</i> - There is RAW data in the run that needs to be reconstructed (create prompt reco jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutPromptRecoMerge</i> - There are RECO data in the run that needs to be merged (create merge jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutAlcaSkim</i> - There is RECO data in the run that needs to be AlcaSkimmed (create AlcaSkim jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">CloseOutAlcaSkimMerge</i> - There are ALCARECO data in the run that needs to be merged (create merge jobs and wait for them to finish)<br />
                <i style="color:blue;font-size:10px">Complete</i> - We have finished processing this run (there could have been failures along the way but we've exhausted all automatic retries)
            """
        ),
        Col(STREAMERS_TITLE, info=
            """The number of streamer files"""
        ),
        Col(REPACKED_TITLE, info=
            """
            The three numbers in the first set represent the number of temporary, final and exported repacked files.<br><br>
            The second set of numbers represent the process to put files in those states, 1 is 100% based on the available files from the previous step.<br><br>
            The coloured bars show graphically the three processing statuses. Red indicates nothing done, green means processing is complete and yellow is somewhere in between. No bar means nothing is currently expected to be done.
            """
        ),
        Col(RECO_TITLE, info=
            """
            The three numbers in the first set represent the number of temporary, final and exported reconstructed files.<br><br>
            The second set of numbers represent the process to put files in those states (1 is 100%) based on the available files from the previous step. This will include files that are not actually processed but acknowledged as such (ie reconstruction is disabled for them).<br><br>
            The coloured bars show graphically the three processing statuses. Red indicates nothing done, green means processing is complete and yellow is somewhere in between. No bar means nothing is currently expected to be done.
            """
        ),
        Col(ALCASKIM_TITLE, info=
            """
            The three numbers in the first set represent the number of temporary, final and exported AlcaSkimmed files.<br><br>
            The second set of numbers represent the process to put files in those states (1 is 100%) based on the available files from the previous step.This will include files that are not actually processed but acknowledged as such (ie AlcaSkim is disabled for them).<br><br>
            The coloured bars show graphically the three processing statuses. Red indicates nothing done, green means processing is complete and yellow is somewhere in between. No bar means nothing is currently expected to be done.
            """
        ),
    ])

    def __init__(self, orderby = 0, asc = True, page = 0, pageSize = 10):
        PagedTableDataProvider.__init__(self, orderby, asc, page, pageSize)
        self.title = "Runs"
        PagedTableDataProvider._load(self)

    def _analysis(self):
        for line in self.data:
            runid = line[self.cols.getId(self.RUN_ID_TITLE)]
            numStreamers = line[self.cols.getId(self.STREAMERS_TITLE)]
            numRepacked = line[self.cols.getId(self.REPACKED_TITLE)]
            numRepackedMerged = line[self.cols.getId(self.REPACKED_TITLE)+1]
            numReco = line[self.cols.getId(self.REPACKED_TITLE)+2]
	    numRecoMerged = line[self.cols.getId(self.REPACKED_TITLE)+3]
            numAlcaSkimmed = line[self.cols.getId(self.REPACKED_TITLE)+4]
	    numAlcaSkimmedMerged = line[self.cols.getId(self.REPACKED_TITLE)+5]
	    numRepackedMergedExported = line[self.cols.getId(self.REPACKED_TITLE)+6]
	    numRecoMergedExported = line[self.cols.getId(self.REPACKED_TITLE)+7]
            numAlcaSkimmedMergedExported = line[self.cols.getId(self.REPACKED_TITLE)+8]
            
	    # analyse repacked percentage
	    repackedP = formatMultiplePercentages([[numRepacked, getRunRepackedPercentage(runid, numStreamers), 'Temp'],[numRepackedMerged, getRunMergePercentage(runid, numRepacked, "RAW"), 'RAW'],[numRepackedMergedExported, getExportedPercentage(numRepackedMerged, numRepackedMergedExported), 'Exp']])  

	    # analyse reco percentage
	    recoP = formatMultiplePercentages([[numReco, getRunProcessedPercentage(runid, "Reconstructable",numReco), 'Temp'],[numRecoMerged, getRunMergePercentage(runid, numReco, "RECO"), 'RECO'],[numRecoMergedExported, getExportedPercentage(numRecoMerged, numRecoMergedExported), 'Exp']])

            # analyse alcaskim percentage
	    alcaSkimmedP = formatMultiplePercentages([[numAlcaSkimmed, getRunProcessedPercentage(runid, "AlcaSkimmable",numAlcaSkimmed), 'Temp'],[numAlcaSkimmedMerged, getRunMergePercentage(runid, numAlcaSkimmed,"ALCARECO"), 'ALCA'],[numAlcaSkimmedMergedExported, getExportedPercentage(numAlcaSkimmedMerged, numAlcaSkimmedMergedExported), 'Exp']])

            line[self.cols.getId(self.REPACKED_TITLE)] = repackedP
            line[self.cols.getId(self.RECO_TITLE)] = recoP
            line[self.cols.getId(self.ALCASKIM_TITLE)] = alcaSkimmedP

	    line = line[:-6]

    def _buildCountQuery(self):
        runTable = Tbl(TABLES.RUN)

        self.countQuery = select(
            [func.count(runTable.c.run_id)]
        )

    def _buildMainQuery(self):
        runTable = Tbl(TABLES.RUN, True)
        runStatusTable = Tbl(TABLES.RUN_STATUS, True)
        streamerTable = Tbl(TABLES.STREAMER, True)
	repackStreamerAssocTable = Tbl(TABLES.REPACK_STREAMER_ASSOC, True)
	wmbsFileDetailsTable = Tbl(TABLES.WMBS_FILE_DETAILS, True)
	wmbsDatasetAssocTable = Tbl(TABLES.WMBS_FILE_DATASET_PATH_ASSOC, True)
	datasetTable = Tbl(TABLES.DATASET_PATH, True)
	dataTierTable = Tbl(TABLES.DATA_TIER, True)
	wmbsFileRunlumiMapTable = Tbl(TABLES.WMBS_FILE_RUNLUMI_MAP, True)
	wmbsSubFilesCompleteTable = Tbl(TABLES.WMBS_SUB_FILES_COMPLETE, True)
	wmbsSubscriptionTable = Tbl(TABLES.WMBS_SUBSCRIPTION, True)
	wmbsSubsTypeTable = Tbl(TABLES.WMBS_SUBS_TYPE, True)
	wmbsFilesetTable = Tbl(TABLES.WMBS_FILESET, True)
	wmbsFilesetFilesTable = Tbl(TABLES.WMBS_FILESET_FILES, True)
	wmbsWorkflowTable = Tbl(TABLES.WMBS_WORKFLOW, True)

        if self.asc: order = "ASC"
        else: order = "DESC"

        # HACK: query time optimisation
        if self.orderby < self.cols.getId(self.STREAMERS_TITLE):
            self.mainQuery = """
            SELECT sub.run_id, sub.start_time, sub.end_time, sub.process, sub.version, sub.expversion, sub.repversion, sub.hltkey, sub.acq_era, sub.status,

            (SELECT COUNT(*) FROM """ + streamerTable + """ s WHERE s.run_id = sub.run_id) numStreamers,

	    (SELECT COUNT(DISTINCT f.id) 
		FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
		WHERE dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'RAW')
                      and  m.run = sub.run_id
                      and fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Mergeable')
		) numRepacked,

            (SELECT COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f 
                        INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
                WHERE dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'RAW')
                      and m.run = sub.run_id
                      and fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'DBSUploadable')
                ) numRepackedMerged,

	    (SELECT COUNT(DISTINCT f.id) 
		FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
		WHERE dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'RECO')
                      and  m.run = sub.run_id
                      and fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Mergeable')
		) numReconstructed,

            (SELECT COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
                WHERE dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'RECO')
                      and m.run = sub.run_id
                      and fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'DBSUploadable')
                ) numReconstructedMerged,

	    (SELECT COUNT(DISTINCT f.id) 
		FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
		WHERE dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'ALCARECO')
                      and  m.run = sub.run_id
                      and fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Mergeable')
		) numAlcaSkimmed,

	    (SELECT COUNT(DISTINCT f.id) 
		FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f.id)
			INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                        INNER JOIN """ + wmbsFileRunlumiMapTable + """ m ON (f.id = m.fileid)
			INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
		WHERE dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'ALCARECO')
                      and  m.run = sub.run_id
                      and fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'DBSUploadable')
                ) numAlcaSkimmedMerged,

            (SELECT  COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
                WHERE fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Deletable')
                       AND f.id IN (SELECT DISTINCT f2.id
                                       FROM """ + wmbsFileDetailsTable + """ f2
                                            INNER JOIN """ + wmbsFileRunlumiMapTable + """ m2 ON (f2.id = m2.fileid)
                                            INNER JOIN """ + wmbsFilesetFilesTable + """ fsf2 ON (f2.id = fsf2.fileid)
                                            INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f2.id)
                                            INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                                       WHERE m2.run = sub.run_id
                                            AND fsf2.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Transferable')
                                            AND dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'RAW')
                                    )
                ) numExportedRepackedMerged,

            (SELECT  COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
                WHERE fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Deletable')
                        AND f.id IN (SELECT DISTINCT f2.id
                                       FROM """ + wmbsFileDetailsTable + """ f2
                                            INNER JOIN """ + wmbsFileRunlumiMapTable + """ m2 ON (f2.id = m2.fileid)
                                            INNER JOIN """ + wmbsFilesetFilesTable + """ fsf2 ON (f2.id = fsf2.fileid)
                                            INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f2.id)
                                            INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                                       WHERE m2.run = sub.run_id
                                            AND fsf2.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Transferable')
                                            AND dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'RECO')
                                     )
                ) numExportedReco,

            (SELECT  COUNT(DISTINCT f.id)
                FROM """ + wmbsFileDetailsTable + """ f
                        INNER JOIN """ + wmbsFilesetFilesTable + """ fsf ON (f.id = fsf.fileid)
                WHERE fsf.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Deletable')
                        AND f.id IN (SELECT DISTINCT f2.id
                                        FROM """ + wmbsFileDetailsTable + """ f2
                                            INNER JOIN """ + wmbsFileRunlumiMapTable + """ m2 ON (f2.id = m2.fileid)
                                            INNER JOIN """ + wmbsFilesetFilesTable + """ fsf2 ON (f2.id = fsf2.fileid)
                                            INNER JOIN """ + wmbsDatasetAssocTable  + """ a ON (a.file_id = f2.id)
                                            INNER JOIN """ + datasetTable  + """ dp ON (a.dataset_path_id = dp.id)
                                        WHERE m2.run = sub.run_id
                                            AND fsf2.fileset = (SELECT id FROM """ + wmbsFilesetTable + """ WHERE name = 'Transferable')
                                            AND dp.data_tier = (SELECT id FROM """ + dataTierTable + """ WHERE name = 'ALCARECO')
                                        )
                ) numExportedAlcaskim

            FROM(
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                (SELECT name FROM cmssw_version WHERE id=r.run_version) version,
                (SELECT name FROM cmssw_version WHERE id=r.express_version) expversion,
                (SELECT name FROM cmssw_version WHERE id=r.repack_version) repversion,
                r.hltkey hltkey,
                r.acq_era acq_era,
                rs.status status
                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order + """
            ) sub"""
        else:
            # slow query

	    #########
	    ## TODO: correct this query completly for the new schema. Is it still useful???
	    #########

            self.mainQuery = """
                SELECT
                r.run_id run_id,
                r.start_time start_time,
                r.end_time end_time,
                r.process process,
                r.version version,
                r.hltkey hltkey,
                r.acq_era acq_era,
                rs.status status,
                (SELECT COUNT(s.run_id) FROM """ + streamerTable + """ s WHERE s.run_id = r.run_id) numStreamers,
                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numRepacked,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numMerged,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numReconstructed,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numExportedRepacked,

                (SELECT COUNT(wfd.id) FROM """ + wmbsFileDetailsTable + """ wfd, """ + streamerTable + """ s, """ + repackStreamerAssocTable + """ rsa WHERE s.run_id = sub.run_id AND s.streamer_id = rsa.streamer_id AND rsa.repacked_id = wfd.id) numExportedReco


                FROM """ + runTable + """ r, """ + runStatusTable + """ rs
                WHERE rs.id = r.run_status
                ORDER BY """ + str(int(self.orderby) + 1) + " " + order

#        globals.logger.log(self.query)
